// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com





#include "bochs.h"





  inline void
bx_cpu_c::write_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length)
{
  Bit32u upper_limit;


  if ( BX_CPU_THIS_PTR protected_mode() ) {
    if ( seg->cache.valid==0 ) {
      bx_printf("seg = %s\n", BX_CPU_THIS_PTR strseg(seg));
      bx_printf("seg->selector.value = %04x\n", (unsigned) seg->selector.value);
      bx_printf("write_virtual_checks: valid bit = 0\n");
bx_printf("CS: %04x\n", (unsigned) BX_CPU_THIS_PTR cs.selector.value);
bx_printf("IP: %04x\n", (unsigned) BX_CPU_THIS_PTR prev_eip);
debug(& BX_CPU_THIS_PTR cs, BX_CPU_THIS_PTR eip);
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }

    if (seg->cache.p == 0) { /* not present */
bx_printf("write_virtual_checks(): segment not present\n");
      exception(int_number(seg), 0, 0);
      return;
      }

    switch ( seg->cache.type ) {
      case 0: case 1:   // read only
      case 4: case 5:   // read only, expand down
      case 8: case 9:   // execute only
      case 10: case 11: // execute/read
      case 12: case 13: // execute only, conforming
      case 14: case 15: // execute/read-only, conforming
bx_printf("write_virtual_checks(): no write access to seg\n");
        exception(int_number(seg), 0, 0);
        return;

      case 2: case 3: /* read/write */
        if ( (offset+length-1) > seg->cache.u.segment.limit_scaled ) {
bx_printf("write_virtual_checks(): write beyond limit, r/w\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;

      case 6: case 7: /* read write, expand down */
        if (seg->cache.u.segment.d_b)
          upper_limit = 0xffffffff;
        else
          upper_limit = 0x0000ffff;
        if ( (offset <= seg->cache.u.segment.limit_scaled) ||
             (offset > upper_limit) ||
             ((upper_limit - offset) < (length - 1)) ) {
bx_printf("write_virtual_checks(): write beyond limit, r/w ED\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;
      }

    return;
    }

  else { /* real mode */
    if ( (offset + length - 1)  >  seg->cache.u.segment.limit_scaled) {
      bx_printf("write_virtual_checks() SEG EXCEPTION:  %x:%x + %x\n",
        (unsigned) seg->selector.value, (unsigned) offset, (unsigned) length);
      if (seg == & BX_CPU_THIS_PTR ss) exception(BX_SS_EXCEPTION, 0, 0);
      else exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    }
}

  inline void
bx_cpu_c::read_virtual_checks(bx_segment_reg_t *seg, Bit32u offset, unsigned length)
{
  Bit32u upper_limit;


  if ( BX_CPU_THIS_PTR protected_mode() ) {
    if ( seg->cache.valid==0 ) {
      bx_printf("seg = %s\n", BX_CPU_THIS_PTR strseg(seg));
      bx_printf("seg->selector.value = %04x\n", (unsigned) seg->selector.value);
      bx_printf("read_virtual_checks: valid bit = 0\n");
bx_printf("CS: %04x\n", (unsigned) BX_CPU_THIS_PTR cs.selector.value);
bx_printf("IP: %04x\n", (unsigned) BX_CPU_THIS_PTR prev_eip);
debug(& BX_CPU_THIS_PTR cs, BX_CPU_THIS_PTR eip);
      exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }

    if (seg->cache.p == 0) { /* not present */
bx_printf("read_virtual_checks(): segment not present\n");
      exception(int_number(seg), 0, 0);
      return;
      }

    switch ( seg->cache.type ) {
      case 0: case 1: /* read only */
      case 10: case 11: /* execute/read */
      case 14: case 15: /* execute/read-only, conforming */
        if ( (offset+length-1) > seg->cache.u.segment.limit_scaled ) {
bx_printf("read_virtual_checks(): write beyond limit\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;

      case 2: case 3: /* read/write */
        if ( (offset+length-1) > seg->cache.u.segment.limit_scaled ) {
bx_printf("read_virtual_checks(): write beyond limit\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;

      case 4: case 5: /* read only, expand down */
        if (seg->cache.u.segment.d_b)
          upper_limit = 0xffffffff;
        else
          upper_limit = 0x0000ffff;
        if ( (offset <= seg->cache.u.segment.limit_scaled) ||
             (offset > upper_limit) ||
             ((upper_limit - offset) < (length - 1)) ) {
bx_printf("read_virtual_checks(): write beyond limit\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;

      case 6: case 7: /* read write, expand down */
        if (seg->cache.u.segment.d_b)
          upper_limit = 0xffffffff;
        else
          upper_limit = 0x0000ffff;
        if ( (offset <= seg->cache.u.segment.limit_scaled) ||
             (offset > upper_limit) ||
             ((upper_limit - offset) < (length - 1)) ) {
bx_printf("read_virtual_checks(): write beyond limit\n");
          exception(int_number(seg), 0, 0);
          return;
          }
        break;

      case 8: case 9: /* execute only */
      case 12: case 13: /* execute only, conforming */
        /* can't read or write an execute-only segment */
bx_printf("read_virtual_checks(): execute only\n");
        exception(int_number(seg), 0, 0);
        return;
        break;
      }
    return;
    }

  else { /* real mode */
    if ( (offset + length - 1)  >  seg->cache.u.segment.limit_scaled) {
      bx_printf("read_virtual_checks() SEG EXCEPTION:  %x:%x + %x\n",
        (unsigned) seg->selector.value, (unsigned) offset, (unsigned) length);
      if (seg == & BX_CPU_THIS_PTR ss) exception(BX_SS_EXCEPTION, 0, 0);
      else exception(BX_GP_EXCEPTION, 0, 0);
      return;
      }
    return;
    }
}




  char *
bx_cpu_c::strseg(bx_segment_reg_t *seg)
{
  if (seg == &BX_CPU_THIS_PTR es) return("ES");
  else if (seg == & BX_CPU_THIS_PTR cs) return("CS");
  else if (seg == & BX_CPU_THIS_PTR ss) return("SS");
  else if (seg == &BX_CPU_THIS_PTR ds) return("DS");
  else if (seg == &BX_CPU_THIS_PTR fs) return("FS");
  else if (seg == &BX_CPU_THIS_PTR gs) return("GS");
  else {
    bx_printf("undefined segment passed to strseg()!\n");
    return("??");
    }
}




  void
bx_cpu_c::write_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data)
{
  write_virtual_checks(seg, offset, 1);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      1, CPL, BX_WRITE, (void *) data);
  else
#endif
    bx_mem.write_physical(seg->cache.u.segment.base + offset, 1, data);
}

  void
bx_cpu_c::write_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data)
{
  write_virtual_checks(seg, offset, 2);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      2, CPL, BX_WRITE, (void *) data);
  else
#endif
    bx_mem.write_physical(seg->cache.u.segment.base + offset, 2, data);
}

  void
bx_cpu_c::write_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data)
{
  write_virtual_checks(seg, offset, 4);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      4, CPL, BX_WRITE, (void *) data);
  else
#endif
    bx_mem.write_physical(seg->cache.u.segment.base + offset, 4, data);
}

  void
bx_cpu_c::read_virtual_byte(bx_segment_reg_t *seg, Bit32u offset, Bit8u *data)
{
  read_virtual_checks(seg, offset, 1);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      1, CPL, BX_READ, (void *) data);
  else
#endif
    bx_mem.read_physical(seg->cache.u.segment.base + offset, 1, data);
}

  void
bx_cpu_c::read_virtual_word(bx_segment_reg_t *seg, Bit32u offset, Bit16u *data)
{
  read_virtual_checks(seg, offset, 2);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      2, CPL, BX_READ, (void *) data);
  else
#endif
    bx_mem.read_physical(seg->cache.u.segment.base + offset, 2, data);
}

  void
bx_cpu_c::read_virtual_dword(bx_segment_reg_t *seg, Bit32u offset, Bit32u *data)
{
  read_virtual_checks(seg, offset, 4);
  
  // all checks OK
#if BX_CPU >= 3
  if (BX_CPU_THIS_PTR cr0.pg)
    access_linear(seg->cache.u.segment.base + offset,
      4, CPL, BX_READ, (void *) data);
  else
#endif
    bx_mem.read_physical(seg->cache.u.segment.base + offset, 4, data);
}
