// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com





#include "bochs.h"




  INLINE void
bx_cpu_c::ADD_EbGb()
{
  Bit8u op2, op1, sum;

#if BX_WEIRDISMS
  {
  Bit8u b1;
  b1 = peek_next_byte();
  if (b1 == 0) {

    bx_printf("add_ebgb(): maybe executing in zero'd memory\n");
    bx_printf("bx_mem.enable_a20=%u\n", (unsigned) bx_mem.enable_a20);
    bx_printf("bx_mem.a20_mask=%08x\n", (unsigned) bx_mem.a20_mask);
    bx_panic("bailing...\n");
    }
  }
#endif

  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, sum);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum);
    }

  SET_FLAGS_OSZAPC_8(op1, op2, sum, BX_INSTR_ADD8);
}


  INLINE void
bx_cpu_c::ADD_EvGv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    /* op2_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, sum_32, BX_INSTR_ADD32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    /* op2_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, sum_16, BX_INSTR_ADD16);
    }
}


  INLINE void
bx_cpu_c::ADD_GbEb()
{
  Bit8u op1, op2, sum;


  decode_exgx();

  /* op1 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op2 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }

  sum = op1 + op2;

  /* now write sum back to destination, which is a register */
  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, sum);

  SET_FLAGS_OSZAPC_8(op1, op2, sum, BX_INSTR_ADD8);
}

  INLINE void
bx_cpu_c::ADD_GvEv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, sum_32;

    /* op1_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, sum_32);

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, sum_32, BX_INSTR_ADD32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;


    /* op1_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    sum_16 = op1_16 + op2_16;
    /* now write sum back to destination */

    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, sum_16);

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, sum_16, BX_INSTR_ADD16);
    }
}

  INLINE void
bx_cpu_c::ADD_ALIb()
{
  Bit8u op1, op2, sum;


  op1 = AL;

  op2 = fetch_next_byte();

  sum = op1 + op2;

  /* now write sum back to destination, which is a register */
  AL = sum;

  SET_FLAGS_OSZAPC_8(op1, op2, sum, BX_INSTR_ADD8);
}

  INLINE void
bx_cpu_c::ADD_eAXIv()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, sum_32;

    op1_32 = EAX;

    op2_32 = fetch_next_dword();

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    EAX = sum_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, sum_32, BX_INSTR_ADD32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;

    op1_16 = AX;

    op2_16 = fetch_next_word();

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    AX = sum_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, sum_16, BX_INSTR_ADD16);
    }
}

  INLINE void
bx_cpu_c::ADC_EbGb()
{
  Bit8u op2, op1, sum;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  sum = op1 + op2 + temp_CF;


  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, sum);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum);
    }

  SET_FLAGS_OSZAPC_8_CF(op1, op2, sum, BX_INSTR_ADC8, temp_CF);
}

  INLINE void
bx_cpu_c::ADC_EvGv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    /* op2_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, sum_32, BX_INSTR_ADC32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    /* op2_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, sum_16, BX_INSTR_ADC16,
                              temp_CF);
    }
}

  INLINE void
bx_cpu_c::ADC_GbEb()
{
  Bit8u op1, op2, sum;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  /* op1 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op2 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2);
    }

  sum = op1 + op2 + temp_CF;

  SET_FLAGS_OSZAPC_8_CF(op1, op2, sum, BX_INSTR_ADC8,
                           temp_CF);

  /* now write sum back to destination, which is a register */
  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, sum);
}

  INLINE void
bx_cpu_c::ADC_GvEv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, sum_32;

    /* op1_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    sum_32 = op1_32 + op2_32 + temp_CF;

    /* now write sum back to destination */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, sum_32);

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, sum_32, BX_INSTR_ADC32,
                             temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;


    /* op1_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    sum_16 = op1_16 + op2_16 + temp_CF;

    /* now write sum back to destination */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, sum_16);

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, sum_16, BX_INSTR_ADC16,
                             temp_CF);
    }
}

  INLINE void
bx_cpu_c::ADC_ALIb()
{
  Bit8u op1, op2, sum;
  Boolean temp_CF;

  temp_CF = get_CF();


  op1 = AL;

  op2 = fetch_next_byte();

  sum = op1 + op2 + temp_CF;

  /* now write sum back to destination, which is a register */
  AL = sum;

  SET_FLAGS_OSZAPC_8_CF(op1, op2, sum, BX_INSTR_ADC8,
                           temp_CF);
}

  INLINE void
bx_cpu_c::ADC_eAXIv()
{
  Boolean temp_CF;

  temp_CF = get_CF();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, sum_32;

    op1_32 = EAX;

    op2_32 = fetch_next_dword();

    sum_32 = op1_32 + op2_32 + temp_CF;

    /* now write sum back to destination */
    EAX = sum_32;

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, sum_32, BX_INSTR_ADC32,
                           temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, sum_16;

    op1_16 = AX;

    op2_16 = fetch_next_word();

    sum_16 = op1_16 + op2_16 + temp_CF;

    /* now write sum back to destination */
    AX = sum_16;

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, sum_16, BX_INSTR_ADC16,
                           temp_CF);
    }
}

  INLINE void
bx_cpu_c::SBB_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - (op2_8 + temp_CF);

  /* now write diff back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_8);
    }

  SET_FLAGS_OSZAPC_8_CF(op1_8, op2_8, diff_8, BX_INSTR_SBB8,
                           temp_CF);
}

  INLINE void
bx_cpu_c::SBB_EvGv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - (op2_32 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, diff_32, BX_INSTR_SBB32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    /* op2_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - (op2_16 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, diff_16, BX_INSTR_SBB16,
                              temp_CF);
    }
}

  INLINE void
bx_cpu_c::SBB_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  /* op1 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op2 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_8);
    }

  diff_8 = op1_8 - (op2_8 + temp_CF);

  /* now write diff back to destination, which is a register */
  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, diff_8);

  SET_FLAGS_OSZAPC_8_CF(op1_8, op2_8, diff_8, BX_INSTR_SBB8,
                           temp_CF);
}

  INLINE void
bx_cpu_c::SBB_GvEv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    diff_32 = op1_32 - (op2_32 + temp_CF);

    /* now write diff back to destination */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, diff_32);

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, diff_32, BX_INSTR_SBB32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;


    /* op1_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    diff_16 = op1_16 - (op2_16 + temp_CF);

    /* now write diff back to destination */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, diff_16);

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, diff_16, BX_INSTR_SBB16,
                              temp_CF);
    }
}

  INLINE void
bx_cpu_c::SBB_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;
  Boolean temp_CF;

  temp_CF = get_CF();


  op1_8 = AL;

  op2_8 = fetch_next_byte();

  diff_8 = op1_8 - (op2_8 + temp_CF);

  /* now write diff back to destination, which is a register */
  AL = diff_8;

  SET_FLAGS_OSZAPC_8_CF(op1_8, op2_8, diff_8, BX_INSTR_SBB8,
                           temp_CF);
}

  INLINE void
bx_cpu_c::SBB_eAXIv()
{
  Boolean temp_CF;

  temp_CF = get_CF();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = fetch_next_dword();

    diff_32 = op1_32 - (op2_32 + temp_CF);

    /* now write diff back to destination */
    EAX = diff_32;

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, diff_32, BX_INSTR_SBB32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;

    op1_16 = AX;

    op2_16 = fetch_next_word();

    diff_16 = op1_16 - (op2_16 + temp_CF);

    /* now write diff back to destination */
    AX = diff_16;

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, diff_16, BX_INSTR_SBB16,
                              temp_CF);
    }
}


  INLINE void
bx_cpu_c::SBB_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  op2_8 = fetch_next_byte();

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - (op2_8 + temp_CF);

  /* now write diff back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_8);
    }

  SET_FLAGS_OSZAPC_8_CF(op1_8, op2_8, diff_8, BX_INSTR_SBB8,
                           temp_CF);
}


  INLINE void
bx_cpu_c::SBB_EvIv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = fetch_next_dword();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - (op2_32 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, diff_32, BX_INSTR_SBB32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = fetch_next_word();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - (op2_16 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, diff_16, BX_INSTR_SBB16,
                              temp_CF);
    }
}

  INLINE void
bx_cpu_c::SBB_EvIb()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) fetch_next_byte();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - (op2_32 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, diff_32, BX_INSTR_SBB32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = (Bit8s) fetch_next_byte();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - (op2_16 + temp_CF);

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, diff_16, BX_INSTR_SBB16,
                              temp_CF);
    }
}



  INLINE void
bx_cpu_c::SUB_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;


  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_8);
    }

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SUB8);
}

  INLINE void
bx_cpu_c::SUB_EvGv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SUB32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    /* op2_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SUB16);
    }
}

  INLINE void
bx_cpu_c::SUB_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;


  decode_exgx();

  /* op1 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op2 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_8);
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination, which is a register */
  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, diff_8);

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SUB8);
}

  INLINE void
bx_cpu_c::SUB_GvEv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR reg_index, diff_32);

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SUB32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;


    /* op1_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR reg_index, diff_16);

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SUB16);
    }
}

  INLINE void
bx_cpu_c::SUB_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;

  op1_8 = AL;

  op2_8 = fetch_next_byte();

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination, which is a register */
  AL = diff_8;

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SUB8);
}

  INLINE void
bx_cpu_c::SUB_eAXIv()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = fetch_next_dword();

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    EAX = diff_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SUB32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;

    op1_16 = AX;

    op2_16 = fetch_next_word();

    diff_16 = op1_16 - op2_16;


    /* now write diff back to destination */
    AX = diff_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SUB16);
    }
}


  INLINE void
bx_cpu_c::CMP_EbGb()
{
  Bit8u op2_8, op1_8, diff_8;


  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - op2_8;

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMP8);
}

  INLINE void
bx_cpu_c::CMP_EvGv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    /* op2_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMP32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    /* op2_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMP16);
    }
}

  INLINE void
bx_cpu_c::CMP_GbEb()
{
  Bit8u op1_8, op2_8, diff_8;

  decode_exgx();

  /* op1 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op2 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op2_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_8);
    }

  diff_8 = op1_8 - op2_8;

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMP8);
}

  INLINE void
bx_cpu_c::CMP_GvEv()
{

  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    /* op1_32 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_32);
      }

    diff_32 = op1_32 - op2_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMP32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;


    /* op1_16 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
    op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR reg_index);

    /* op2_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op2_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op2_16);
      }

    diff_16 = op1_16 - op2_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMP16);
    }
}


  INLINE void
bx_cpu_c::CMP_ALIb()
{
  Bit8u op1_8, op2_8, diff_8;


  op1_8 = AL;

  op2_8 = fetch_next_byte();

  diff_8 = op1_8 - op2_8;

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMP8);
}

  INLINE void
bx_cpu_c::CMP_eAXIv()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, op2_32, diff_32;

    op1_32 = EAX;

    op2_32 = fetch_next_dword();

    diff_32 = op1_32 - op2_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMP32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, op2_16, diff_16;

    op1_16 = AX;

    op2_16 = fetch_next_word();

    diff_16 = op1_16 - op2_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMP16);
    }
}

  INLINE void
bx_cpu_c::INC_eAX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u eax;

    eax = EAX;

    eax++;

    EAX = eax;

    SET_FLAGS_OSZAP_32(0, 0, eax, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = AX;

    ax++;

    AX = ax;

    SET_FLAGS_OSZAP_16(0, 0, ax, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eCX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ecx;

    ecx = ECX;

    ecx++;

    ECX = ecx;

    SET_FLAGS_OSZAP_32(0, 0, ecx, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u cx;

    cx = CX;

    cx++;

    CX = cx;

    SET_FLAGS_OSZAP_16(0, 0, cx, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eDX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edx;

    edx = EDX;

    edx++;

    EDX = edx;

    SET_FLAGS_OSZAP_32(0, 0, edx, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u dx;

    dx = DX;

    dx++;

    DX = dx;

    SET_FLAGS_OSZAP_16(0, 0, dx, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eBX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebx;

    ebx = EBX;

    ebx++;

    EBX = ebx;

    SET_FLAGS_OSZAP_32(0, 0, ebx, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bx;

    bx = BX;

    bx++;

    BX = bx;

    SET_FLAGS_OSZAP_16(0, 0, bx, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eSP()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esp;

    esp = ESP;

    esp++;

    ESP = esp;

    SET_FLAGS_OSZAP_32(0, 0, esp, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u sp;

    sp = SP;

    sp++;

    SP = sp;

    SET_FLAGS_OSZAP_16(0, 0, sp, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eBP()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebp;

    ebp = EBP;

    ebp++;

    EBP = ebp;

    SET_FLAGS_OSZAP_32(0, 0, ebp, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bp;

    bp = BP;

    bp++;

    BP = bp;

    SET_FLAGS_OSZAP_16(0, 0, bp, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eSI()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esi;

    esi = ESI;

    esi++;

    ESI = esi;

    SET_FLAGS_OSZAP_32(0, 0, esi, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u si;

    si = SI;

    si++;

    SI = si;

    SET_FLAGS_OSZAP_16(0, 0, si, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::INC_eDI()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edi;

    edi = EDI;

    edi++;

    EDI = edi;

    SET_FLAGS_OSZAP_32(0, 0, edi, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u di;

    di = DI;

    di++;

    DI = di;

    SET_FLAGS_OSZAP_16(0, 0, di, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eAX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u eax;

    eax = EAX;

    eax--;

    EAX = eax;

    SET_FLAGS_OSZAP_32(0, 0, eax, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u ax;

    ax = AX;

    ax--;

    AX = ax;

    SET_FLAGS_OSZAP_16(0, 0, ax, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eCX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ecx;

    ecx = ECX;

    ecx--;

    ECX = ecx;

    SET_FLAGS_OSZAP_32(0, 0, ecx, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u cx;

    cx = CX;

    cx--;

    CX = cx;

    SET_FLAGS_OSZAP_16(0, 0, cx, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eDX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edx;

    edx = EDX;

    edx--;

    EDX = edx;

    SET_FLAGS_OSZAP_32(0, 0, edx, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u dx;

    dx = DX;

    dx--;

    DX = dx;

    SET_FLAGS_OSZAP_16(0, 0, dx, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eBX()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebx;

    ebx = EBX;

    ebx--;

    EBX = ebx;

    SET_FLAGS_OSZAP_32(0, 0, ebx, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bx;

    bx = BX;

    bx--;

    BX = bx;

    SET_FLAGS_OSZAP_16(0, 0, bx, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eSP()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esp;

    esp = ESP;

    esp--;

    ESP = esp;

    SET_FLAGS_OSZAP_32(0, 0, esp, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u sp;

    sp = SP;

    sp--;

    SP = sp;

    SET_FLAGS_OSZAP_16(0, 0, sp, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eBP()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u ebp;

    ebp = EBP;

    ebp--;

    EBP = ebp;

    SET_FLAGS_OSZAP_32(0, 0, ebp, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u bp;

    bp = BP;

    bp--;

    BP = bp;

    SET_FLAGS_OSZAP_16(0, 0, bp, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eSI()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u esi;

    esi = ESI;

    esi--;

    ESI = esi;

    SET_FLAGS_OSZAP_32(0, 0, esi, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u si;

    si = SI;

    si--;

    SI = si;

    SET_FLAGS_OSZAP_16(0, 0, si, BX_INSTR_DEC16);
    }
}

  INLINE void
bx_cpu_c::DEC_eDI()
{
#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u edi;

    edi = EDI;

    edi--;

    EDI = edi;

    SET_FLAGS_OSZAP_32(0, 0, edi, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u di;

    di = DI;

    di--;

    DI = di;

    SET_FLAGS_OSZAP_16(0, 0, di, BX_INSTR_DEC16);
    }
}


  INLINE void
bx_cpu_c::CBW()
{
  /* CBW: no flags are effected */

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* CWDE opcode */
    EAX = (Bit16s) AX;
    }
  else
#endif /* BX_CPU > 2 */
    { /* CBW opcode */
    AX = (Bit8s) AL;
    }
}

  INLINE void
bx_cpu_c::CWD()
{
  /* CWD: no flags are affected */

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* CDQ opcode */
    if (EAX & 0x80000000) {
      EDX = 0xFFFFFFFF;
      }
    else {
      EDX = 0x00000000;
      }
    }
  else
#endif /* BX_CPU > 2 */
    { /* CWD opcode */
    if (AX & 0x8000) {
      DX = 0xFFFF;
      }
    else {
      DX = 0x0000;
      }
    }
}

  INLINE void
bx_cpu_c::CMPXCHG_XBTS()
{
#if BX_CPU < 4
  bx_printf("CMPXCHG_XBTS:\n");
  exception(6, 0, 0);
  return;
#else
  bx_panic("CMPXCHG_XBTS:\n");
#endif
}

  INLINE void
bx_cpu_c::CMPXCHG_IBTS()
{
  bx_panic("CMPXCHG_IBTS:\n");
}

  INLINE void
bx_cpu_c::XADD_EbGb()
{
#if BX_CPU < 4
  bx_panic("XADD_EbGb: not supported on < 80486\n");
#else

  Bit8u op2, op1, sum;
  Bit8u op2_b7, op1_b7, sum_b7;

  /* XADD dst(r/m8), src(r8)
   * temp <-- src + dst         | sum = op2 + op1
   * src  <-- dst               | op2 = op1
   * dst  <-- tmp               | op1 = sum
   */

  decode_exgx();

  /* op2 is a register, BX_CPU_THIS_PTR rm_addr is an index of a register */
  op2 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR reg_index);

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, sum);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum);
    }

  /* and write destination into source */
  BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR reg_index, op1);

  SET_FLAGS_OSZAPC_8(op1, op2, sum, BX_INSTR_XADD8);
#endif
}

  INLINE void
bx_cpu_c::XADD_EvGv()
{
  bx_panic("XADD_EvGv:\n");
}

  INLINE void
bx_cpu_c::ADD_EbIb()
{
  Bit8u op2, op1, sum;


  /* op1 is mod+r/m, op2 is an 8bit immediate */
  decode_exgx();

  op2 = fetch_next_byte();

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  sum = op1 + op2;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, sum);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum);
    }

  SET_FLAGS_OSZAPC_8(op1, op2, sum, BX_INSTR_ADD8);
}

  INLINE void
bx_cpu_c::ADC_EbIb()
{
  Bit8u op2, op1, sum;
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

  op2 = fetch_next_byte();

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  sum = op1 + op2 + temp_CF;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, sum);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum);
    }

  SET_FLAGS_OSZAPC_8_CF(op1, op2, sum, BX_INSTR_ADC8,
                           temp_CF);
}


  INLINE void
bx_cpu_c::SUB_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;


  decode_exgx();

  op2_8 = fetch_next_byte();

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - op2_8;

  /* now write diff back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_8);
    }

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_SUB8);
}

  INLINE void
bx_cpu_c::CMP_EbIb()
{
  Bit8u op2_8, op1_8, diff_8;


  decode_exgx();

  op2_8 = fetch_next_byte();

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = op1_8 - op2_8;

  SET_FLAGS_OSZAPC_8(op1_8, op2_8, diff_8, BX_INSTR_CMP8);
}


  INLINE void
bx_cpu_c::ADD_EvIv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = fetch_next_dword();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, sum_32, BX_INSTR_ADD32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    op2_16 = fetch_next_word();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, sum_16, BX_INSTR_ADD16);
    }
}

  INLINE void
bx_cpu_c::ADC_EvIv()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = fetch_next_dword();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, sum_32, BX_INSTR_ADC32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    op2_16 = fetch_next_word();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, sum_16, BX_INSTR_ADC16,
                              temp_CF);
    }
}


  INLINE void
bx_cpu_c::SUB_EvIv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = fetch_next_dword();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SUB32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = fetch_next_word();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SUB16);
    }
}

  INLINE void
bx_cpu_c::CMP_EvIv()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = fetch_next_dword();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMP32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = fetch_next_word();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMP16);
    }
}

  INLINE void
bx_cpu_c::ADD_EvIb()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = (Bit8s) fetch_next_byte();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, sum_32, BX_INSTR_ADD32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    op2_16 = (Bit8s) fetch_next_byte();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, sum_16, BX_INSTR_ADD16);
    }
}

  INLINE void
bx_cpu_c::ADC_EvIb()
{
  Boolean temp_CF;

  temp_CF = get_CF();


  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, sum_32;

    op2_32 = (Bit8s) fetch_next_byte();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    sum_32 = op1_32 + op2_32 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_32);
      }

    SET_FLAGS_OSZAPC_32_CF(op1_32, op2_32, sum_32, BX_INSTR_ADC32,
                              temp_CF);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, sum_16;


    op2_16 = (Bit8s) fetch_next_byte();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    sum_16 = op1_16 + op2_16 + temp_CF;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, sum_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &sum_16);
      }

    SET_FLAGS_OSZAPC_16_CF(op1_16, op2_16, sum_16, BX_INSTR_ADC16,
                              temp_CF);
    }
}

  INLINE void
bx_cpu_c::SUB_EvIb()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) fetch_next_byte();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_SUB32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = (Bit8s) fetch_next_byte();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_SUB16);
    }
}

  INLINE void
bx_cpu_c::CMP_EvIb()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op2_32, op1_32, diff_32;

    op2_32 = (Bit8s) fetch_next_byte();

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = op1_32 - op2_32;

    SET_FLAGS_OSZAPC_32(op1_32, op2_32, diff_32, BX_INSTR_CMP32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op2_16, op1_16, diff_16;


    op2_16 = (Bit8s) fetch_next_byte();

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = op1_16 - op2_16;

    SET_FLAGS_OSZAPC_16(op1_16, op2_16, diff_16, BX_INSTR_CMP16);
    }
}


  INLINE void
bx_cpu_c::NEG_Eb()
{
  Bit8u op1_8, diff_8;

  decode_exgx();

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  diff_8 = 0 - op1_8;

  /* now write diff back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_8);
    }

  SET_FLAGS_OSZAPC_8(op1_8, 0, diff_8, BX_INSTR_NEG8);
}

  INLINE void
bx_cpu_c::NEG_Ev()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    /* for 32 bit operand size mode */
    Bit32u op1_32, diff_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    diff_32 = 0 - op1_32;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_32);
      }

    SET_FLAGS_OSZAPC_32(op1_32, 0, diff_32, BX_INSTR_NEG32);
    }
  else
#endif /* BX_CPU > 2 */
    { /* 16 bit operand size mode */
    Bit16u op1_16, diff_16;


    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    diff_16 = 0 - op1_16;

    /* now write diff back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, diff_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &diff_16);
      }

    SET_FLAGS_OSZAPC_16(op1_16, 0, diff_16, BX_INSTR_NEG16);
    }
}


  INLINE void
bx_cpu_c::INC_Eb()
{
  Bit8u  op1;

  decode_exgx();

  /* op1 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }


  op1++;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, op1);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1);
    }

  SET_FLAGS_OSZAP_8(0, 0, op1, BX_INSTR_INC8);
}

  INLINE void
bx_cpu_c::INC_Ev()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    op1_32++;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    SET_FLAGS_OSZAP_32(0, 0, op1_32, BX_INSTR_INC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    op1_16++;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    SET_FLAGS_OSZAP_16(0, 0, op1_16, BX_INSTR_INC16);
    }
}

  INLINE void
bx_cpu_c::DEC_Eb()
{
  Bit8u op1_8;

  decode_exgx();

  /* op1_8 is a register or memory reference */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    op1_8 = BX_READ_8BIT_REG(BX_CPU_THIS_PTR rm_addr);
    }
  else {
    /* pointer, segment address pair */
    read_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  op1_8--;

  /* now write sum back to destination */
  if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
    BX_WRITE_8BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_8);
    }
  else {
    write_virtual_byte(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_8);
    }

  SET_FLAGS_OSZAP_8(0, 0, op1_8, BX_INSTR_DEC8);
}

  INLINE void
bx_cpu_c::DEC_Ev()
{
  decode_exgx();

#if BX_CPU > 2
  if (BX_CPU_THIS_PTR is_32bit_opsize) { /* 32 bit operand size mode */
    Bit32u op1_32;

    /* op1_32 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_32 = BX_READ_32BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    op1_32--;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_32BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_32);
      }
    else {
      write_virtual_dword(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_32);
      }

    SET_FLAGS_OSZAP_32(0, 0, op1_32, BX_INSTR_DEC32);
    }
  else
#endif /* BX_CPU > 2 */
    {
    Bit16u op1_16;

    /* op1_16 is a register or memory reference */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      op1_16 = BX_READ_16BIT_REG(BX_CPU_THIS_PTR rm_addr);
      }
    else {
      /* pointer, segment address pair */
      read_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    op1_16--;

    /* now write sum back to destination */
    if (BX_CPU_THIS_PTR rm_type == BX_REGISTER_REF) {
      BX_WRITE_16BIT_REG(BX_CPU_THIS_PTR rm_addr, op1_16);
      }
    else {
      write_virtual_word(BX_CPU_THIS_PTR rm_seg_reg, BX_CPU_THIS_PTR rm_addr, &op1_16);
      }

    SET_FLAGS_OSZAP_16(0, 0, op1_16, BX_INSTR_DEC16);
    }
}

INLINE void bx_cpu_c::CMPXCHG_EbGb() {bx_panic("CMPXCHG_EbGb:\n");}
INLINE void bx_cpu_c::CMPXCHG_EvGv() {bx_panic("CMPXCHG_EvGv:\n");}
