// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com


#ifndef  __cplusplus
// strict C programs at least need Bit8u, Bit16u, Bit32u, Boolean typedefs
#include "config.h"
#endif


#define BX_MAX_PATH     256
 
#ifdef  __cplusplus
extern "C" {
#endif


// Flex defs
extern int yylex(void);
extern char *yytext;  // Using the pointer option rather than array
extern int yywrap(void);

// Yacc defs
extern int yyparse(void);
extern void yyerror(char *s);

void bx_dbg_continue(void);
void bx_dbg_single_step(void);
void bx_dbg_set(char *p1, char *p2, char *p3);
void bx_dbg_breakpoint(char *p1, char *p2);
void bx_dbg_del_breakpoint(unsigned handle);
void bx_dbg_exit(int code);
void bx_dbg_vbreakpoint(Boolean specific, Bit32u cs, Bit32u eip);
void bx_dbg_lbreakpoint(Boolean specific, Bit32u laddress);
void bx_dbg_pbreakpoint(Boolean specific, Bit32u paddress);
void bx_dbg_info_bpoints(void);

Boolean bx_dbg_cpu_icount(Bit32u cs, Bit32u eip,
                       Bit32u laddr, Bit32u paddr, Boolean paddr_valid,
                       Boolean intr_pending, Boolean is_32bit_code);

extern int bx_dbg_extensions(char *command);
extern void bx_add_lex_input(char *buf);


typedef struct {
  Bit16u  cs;
  Bit32u  eip;
  Bit32u  linear_addr;
  Bit32u  phy_addr;
  Boolean phy_addr_valid;
  } bx_instr_addr_t;
 
typedef struct {
  Boolean          event_occurred; // any break condition will trip this

  struct {
    Boolean          intr_pending;
    bx_instr_addr_t  iaddr;
    Boolean          is_32bit_code;
    } before;

  struct {
    Boolean          intr_pending;
    bx_instr_addr_t  iaddr;
    Boolean          is_32bit_code;
    } after;
 
  Boolean          is_instr_count_bpoint;
  Bit32u              instr_count;

  Boolean          is_vir_iaddr_bpoint;
  unsigned            vir_iaddr_handle;
 
  Boolean          is_lin_iaddr_bpoint;
  unsigned            lin_iaddr_handle;
 
  Boolean          is_phy_iaddr_bpoint;
  unsigned            phy_iaddr_handle;

  Boolean          is_io_bpoint;
  unsigned            io_handle;
  unsigned            io_addr;
 
  Boolean          is_soft_int_bpoint;
  Boolean             soft_int_handle;
  Boolean             soft_int_num;
 
  Boolean          is_hard_int_bpoint;
  Boolean             hard_int_handle;
  Boolean             hard_int_num;
 
  Boolean          is_memory_bpoint; // do later
 
  Boolean          is_opcode_bpoint; // do later
 
  Boolean          is_exception_bpoint; // do later
 
  // DMA stuff // do later
 
  Boolean          is_user_interrupt;
  } bx_bpoint_state_t;



enum bx_disposition_e { bx_disp_disable=0, bx_disp_keep=1, bx_disp_delete=2 };

typedef struct {
  enum bx_disposition_e disposition;
  Boolean               enabled;
  unsigned long         ignorecount;
  } bx_bpoint_attr_t;


#define BX_MAX_IADDR_BPOINTS 10

#define BX_BPOINT_TYPE_NONE     0
#define BX_BPOINT_TYPE_VIRTUAL  1
#define BX_BPOINT_TYPE_LINEAR   2
#define BX_BPOINT_TYPE_PHYSICAL 3

typedef struct {
  Bit32u instr_count_max;

  unsigned num_iaddr_bpoints;

  unsigned bpoint_type[BX_MAX_IADDR_BPOINTS];
  bx_bpoint_attr_t attr[BX_MAX_IADDR_BPOINTS];
  union {
    struct {
      Bit32u           cs; // really only use 16 bits
      Bit32u           eip;
      } vir;

    struct {
      Bit32u           addr;
      } lin;

    struct {
      Bit32u           addr;
      } phy;
    } iaddr[BX_MAX_IADDR_BPOINTS];

  } bx_bpoint_requests_t;



extern volatile Boolean bx_dbg_interrupt_requested;
extern bx_bpoint_requests_t bx_dbg_bpoint_requests;

#ifdef  __cplusplus
}
#endif

extern int bx_dbg_main(int argc, char *argv[]);
