%{
#include <stdio.h>
#include <stdlib.h>
#include "debug.h"
// %left '-' '+'
// %left '*' '/'
// %right
// %nonassoc UMINUS
%}


%union {
  char    *sval;
  Bit32u   uval;
  }

%token <sval> BX_TOKEN_CONTINUE
%token <sval> BX_TOKEN_SINGLE_STEP
%token <sval> BX_TOKEN_NEXT_STEP
%token <sval> BX_TOKEN_SET
%token <sval> BX_TOKEN_VBREAKPOINT
%token <sval> BX_TOKEN_LBREAKPOINT
%token <sval> BX_TOKEN_PBREAKPOINT
%token <sval> BX_TOKEN_DEL_BREAKPOINT
%token <sval> BX_TOKEN_INFO

%token <sval> BX_TOKEN_DIS
%token <sval> BX_TOKEN_ON
%token <sval> BX_TOKEN_OFF

%token <uval> BX_TOKEN_NUMERIC

%start command_list

%%
command_list:
      command '\n'                 { }
    | '\n'                         { }
    ;

command:
      BX_TOKEN_CONTINUE
        {
        bx_dbg_continue();
        free($1);
        }
    | BX_TOKEN_SINGLE_STEP
        {
        bx_dbg_single_step();
        free($1);
        }
    | BX_TOKEN_NEXT_STEP
        {
        bx_dbg_single_step();
        free($1);
        }
    | set_command
    | breakpoint_command
    | info_command
    | delete_command
    ;

set_command:
      BX_TOKEN_SET BX_TOKEN_DIS BX_TOKEN_ON
        {
        bx_dbg_set($1, $2, $3);
        free($1); free($2); free($3);
        }
    | BX_TOKEN_SET BX_TOKEN_DIS BX_TOKEN_OFF
        {
        bx_dbg_set($1, $2, $3);
        free($1); free($2); free($3);
        }
    ;

breakpoint_command:
      BX_TOKEN_VBREAKPOINT
        {
        bx_dbg_vbreakpoint(0, 0, 0);
        free($1);
        }
    | BX_TOKEN_VBREAKPOINT BX_TOKEN_NUMERIC ':' BX_TOKEN_NUMERIC
        {
        bx_dbg_vbreakpoint(1, $2, $4);
        free($1);
        }
    | BX_TOKEN_LBREAKPOINT
        {
        bx_dbg_lbreakpoint(0, 0);
        free($1);
        }
    | BX_TOKEN_LBREAKPOINT BX_TOKEN_NUMERIC
        {
        bx_dbg_lbreakpoint(1, $2);
        free($1);
        }
    | BX_TOKEN_PBREAKPOINT
        {
        bx_dbg_pbreakpoint(0, 0);
        free($1);
        }
    | BX_TOKEN_PBREAKPOINT BX_TOKEN_NUMERIC
        {
        bx_dbg_pbreakpoint(1, $2);
        free($1);
        }
    ;

info_command:
      BX_TOKEN_INFO BX_TOKEN_PBREAKPOINT
        {
        bx_dbg_info_bpoints();
        free($1); free($2);
        }
    ;

delete_command:
      BX_TOKEN_DEL_BREAKPOINT BX_TOKEN_NUMERIC
        {
        bx_dbg_del_breakpoint($2);
        free($1);
        }
    ;

%%

