// Copyright Notice
// ================
// BOCHS is Copyright 1994,1995,1996,1997,1998 by Kevin P. Lawton.
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02154
// 
//     EMail:    bochs@world.std.com







#include "bochs.h"
extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
}





#if BX_PROVIDE_CPU_MEMORY
  Bit32u
bx_memory_c::get_memory_in_k(void)
{
fprintf(stderr, "get_memory_in_k() = %u\n", (unsigned)
  (megabytes * 1024));

  return(megabytes * 1024);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  // bx_memory_c constructor
bx_memory_c::bx_memory_c(void)
{
  bx_mem.vector = NULL;
  bx_mem.len    = 0;
  bx_mem.megabytes = 0;
}
#endif // #if BX_PROVIDE_CPU_MEMORY



#if BX_PROVIDE_CPU_MEMORY
  // bx_memory_c constructor
bx_memory_c::bx_memory_c(size_t memsize)
{
  fprintf(stderr, "bx_memory_c::bx_memory_c(size_t) called\n");
  fprintf(stderr, "  memsize = %u\n", (unsigned) memsize);
  vector = new Bit8u[memsize];
  len    = memsize;
  megabytes = len / (1024*1024);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
// bx_memory_c destructor
bx_memory_c::~bx_memory_c(void)
{
  if (this-> vector != NULL) {
    delete this->vector;
    }
  else {
    fprintf(stderr, "  memory not freed as it wasn't allocated!\n");
    }
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  void
bx_memory_c::init_memory(int memsize)
{
  // you can pass 0 if memory has been allocated already through
  // the constructor, or the desired size of memory if it hasn't

  if (this->vector == NULL) {
    // memory not already allocated, do now...
    fprintf(stderr, "bx_memory_c::init_memory(int): allocating memory.\n");
    fprintf(stderr, "  memsize = %u\n", (unsigned) memsize);
    vector = new Bit8u[memsize];
    len    = memsize;
    megabytes = memsize / (1024*1024);
    }
  memset(this->vector, 0x00, len);
}
#endif // #if BX_PROVIDE_CPU_MEMORY


#if BX_PROVIDE_CPU_MEMORY
  void
bx_memory_c::load_ROM(const char *path, Bit32u romaddress)
{
  struct stat stat_buf;
  int fd, ret;
  unsigned long size, offset;

  // read in ROM BIOS image file
  fd = open(path, O_RDONLY
#ifdef O_BINARY
            | O_BINARY
#endif
           );
  if (fd < 0) {
    fprintf(stderr, "load_ROM: couldn't open ROM image file '%s'.\n", path);
    exit(1);
    }
  ret = fstat(fd, &stat_buf);
  if (ret) {
    fprintf(stderr, "load_ROM: couldn't stat ROM image file '%s'.\n", path);
    exit(1);
    }

  size = stat_buf.st_size;

  if ( (romaddress + size) > this->len ) {
    fprintf(stderr, "load_ROM: ROM address range > physical memsize!\n");
    exit(1);
    }

  offset = 0;
  while (size > 0) {
    ret = read(fd, (void *) &this->vector[romaddress + offset], size);
    if (ret <= 0) {
      fprintf(stderr, "load_ROM: read failed on ROM BIOS image\n");
      exit(1);
      }
    size -= ret;
    offset += ret;
    }
  close(fd);
  fprintf(stderr, "load_ROM: ROM BIOS '%s', size=%u read into memory at %08x\n",
          path, (unsigned) stat_buf.st_size, (unsigned) romaddress);
}
#endif // #if BX_PROVIDE_CPU_MEMORY



#if BX_PROVIDE_CPU_MEMORY

  void
bx_memory_c::write_physical(Bit32u addr, unsigned len, void *data)
{
  Bit32u i;
  Bit8u  *data_ptr;

  addr = A20ADDR(addr);

  BX_MEMORY_HOOK (addr, len, BX_WRITE, data);

  if ( (addr + len) > this->len ) {
    return;
#if 0
    bx_panic("access beyond physical memory boundary! (%08x+%02x)\n",
      (unsigned) addr, (unsigned) len);
#endif
    }


#ifdef BX_LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;

  for (i=1; i<=len; i++, addr++, data_ptr++)
#else // BX_BIG_ENDIAN
  data_ptr = (Bit8u *) data + sizeof(Bit8u) * (len - 1);

  for (i=1; i<=len; i++, addr++, data_ptr--)
#endif
    {
    if ( (addr & 0xFFFC0000)==0x000C0000 ) {
      bx_printf("ROM lock %08x: len=%u\n",
        (unsigned) addr, (unsigned) len);
      return;
      }

#if BX_SUPPORT_VGA
    if ( (addr & 0xfffe0000) == 0x000a0000 )
      bx_devices.vga->mem_write(addr, *(Bit8u *) data_ptr);
    else
#endif
      bx_mem.vector[addr] = *(Bit8u *) data_ptr;
    } /* for (i=... */
}


  void
bx_memory_c::read_physical(Bit32u addr, unsigned len, void *data)
{
  Bit32u i;
  Bit8u  *data_ptr;

  addr = A20ADDR(addr);

  BX_MEMORY_HOOK (addr, len, BX_READ, data);

  if ( (addr + len) > this->len ) {
    for (i = 0; i < len; i++)
      ( (Bit8u *) data)[i] = 0xff;
    return;
#if 0
    bx_panic("access beyond physical memory boundary! (%08x+%02x)\n",
      (unsigned) addr, (unsigned) len);
#endif
    }


#ifdef BX_LITTLE_ENDIAN
  data_ptr = (Bit8u *) data;

  for (i=1; i<=len; i++, addr++, data_ptr++)
#else // BX_BIG_ENDIAN
  data_ptr = (Bit8u *) data + sizeof(Bit8u) * (len - 1);

  for (i=1; i<=len; i++, addr++, data_ptr--)
#endif
    {
#if BX_SUPPORT_VGA
    if ( (addr & 0xfffe0000) == 0x000a0000 )
      *(Bit8u *) data_ptr = bx_devices.vga->mem_read(addr);
    else
#endif
      *(Bit8u *) data_ptr = bx_mem.vector[addr];
    } /* for (i=... */
}




#endif // #if BX_PROVIDE_CPU_MEMORY


  void
bx_dbg_fetch_mem(Bit32u addr, unsigned len, Bit8u *buf)
{
  memcpy(buf, &bx_mem.vector[addr], len);
}
